# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module is responsible for launching protocol executions.
There are two main scenarios: local execution and remote execution.

A. Local execution: 
This will depend on the 'localhost' configuration
1- Check if the protocol will be launched with MPI or not (using MPI template from config)
2- Check if the protocol will be submitted to a queue (using Queue template from config)
3- Build the command that will be launched.

B. Remote execution:
1- Establish a connection with remote host for protocol execution
2- Copy necessary files to remote host.
3- Run a local process (for local execution, see case A) in the remote host
4- Get the result back after launching remotely
"""

import os
import re
from subprocess import Popen, PIPE
import pyworkflow as pw
from pyworkflow.exceptions import PyworkflowException
from pyworkflow.utils import (redStr, greenStr, makeFilePath, join, process,
                              getHostFullName)
from pyworkflow.protocol.constants import UNKNOWN_JOBID

LOCALHOST = 'localhost'


# ******************************************************************
# *         Public functions provided by the module
# ******************************************************************

def launch(protocol, wait=False, stdin=None, stdout=None, stderr=None):
    """ This function should be used to launch a protocol
    This function will decide which case, A or B will be used.
    """
    if _isLocal(protocol):
        jobId = _launchLocal(protocol, wait, stdin, stdout, stderr)
    else:
        jobId = _launchRemote(protocol, wait)

    protocol.setJobId(jobId)

    return jobId


def stop(protocol):
    """ 
    """
    if _isLocal(protocol):
        return _stopLocal(protocol)
    else:
        return _stopRemote(protocol)


def schedule(protocol, initialSleepTime=0, wait=False):
    """ Use this function to schedule protocols that are not ready to
    run yet. Right now it only make sense to schedule jobs locally.
    """
    cmd = '%s %s' % (pw.PYTHON, pw.getScheduleScript())
    cmd += ' "%s" "%s" %s --initial_sleep %s' % (protocol.getProject().path,
                              protocol.getDbPath(),
                              protocol.strId(), initialSleepTime)
    jobId = _run(cmd, wait)
    protocol.setJobId(jobId)

    return jobId


# ******************************************************************
# *         Internal utility functions
# ******************************************************************
def _isLocal(protocol):
    return protocol.getHostName() == LOCALHOST


def _runsLocally(protocol):
    """ Return True if this protocol is running in this machine,
    where the PID makes sense.
    """
    return protocol.getHostFullName() == getHostFullName()


# ******************************************************************
# *                 Function related to LAUNCH
# ******************************************************************
def _getAppsProgram(prog):
    """ Get a command to launch a program under the apps folder.
    """
    return "%s %s" % (pw.PYTHON, pw.join(pw.APPS, prog))


def _launchLocal(protocol, wait, stdin=None, stdout=None, stderr=None):
    # Check first if we need to launch with MPI or not
    command = ('%s %s "%s" "%s" %s'
               % (pw.PYTHON, pw.join(pw.APPS, 'pw_protocol_run.py'),
                  protocol.getProject().path, protocol.getDbPath(),
                  protocol.strId()))
    hostConfig = protocol.getHostConfig()
    useQueue = protocol.useQueue()
    # Check if need to submit to queue    
    if useQueue and (protocol.getSubmitDict()["QUEUE_FOR_JOBS"] == "N"):
        submitDict = dict(hostConfig.getQueuesDefault())
        submitDict.update(protocol.getSubmitDict())
        submitDict['JOB_COMMAND'] = command
        jobId = _submit(hostConfig, submitDict)
    else:
        jobId = _run(command, wait, stdin, stdout, stderr)

    return jobId


def _runRemote(protocol, mode):
    """ Launch remotely 'pw_protocol_remote.py' script to run or stop a protocol. 
    Params:
        protocol: the protocol to be ran or stopped.
        mode: should be either 'run' or 'stop'
    """
    host = protocol.getHostConfig()
    tpl = "ssh %(address)s '%(scipion)s/scipion "

    if host.getScipionConfig() is not None:
        tpl += "--config %(config)s "

    tpl += "runprotocol pw_protocol_remote.py %(mode)s "
    tpl += "%(project)s %(protDb)s %(protId)s' "

    # Use project base name,
    # in remote SCIPION_USER_DATA/projects should be prepended
    projectPath = os.path.basename(protocol.getProject().path)

    args = {'address': host.getAddress(),
            'mode': mode,
            'scipion': host.getScipionHome(),
            'config': host.getScipionConfig(),
            'project': projectPath,
            'protDb': protocol.getDbPath(),
            'protId': protocol.getObjId()
            }
    cmd = tpl % args
    print("** Running remote: %s" % greenStr(cmd))
    p = Popen(cmd, shell=True, stdout=PIPE)

    return p


def _launchRemote(protocol, wait):
    p = _runRemote(protocol, 'run')
    jobId = UNKNOWN_JOBID
    out, err = p.communicate()
    if err:
        raise Exception(err)
    s = re.search('Scipion remote jobid: (\d+)', out)
    if s:
        jobId = int(s.group(1))
    else:
        raise Exception("** Couldn't parse ouput: %s" % redStr(out))

    return jobId


def _copyFiles(protocol, rpath):
    """ Copy all required files for protocol to run
    in a remote execution host.
    NOTE: this function should always be execute with 
    the current working dir pointing to the project dir.
    And the remotePath is assumed to be in protocol.getHostConfig().getHostPath()
    Params:
        protocol: protocol to copy files
        ssh: an ssh connection to copy the files.
    """
    remotePath = protocol.getHostConfig().getHostPath()

    for f in protocol.getFiles():
        remoteFile = join(remotePath, f)
        rpath.putFile(f, remoteFile)


def analyzeFormattingTypeError(string, dictionary):
    """ receives a string with %(VARS) to be replaced with a dictionary
     it splits te string by \n and test the formatting per line. Raises an exception if any line fails
     with all problems found"""

    # Do the replace line by line
    lines = string.split("\n")

    problematicLines = []
    for line in lines:
        try:
            line % dictionary
        except KeyError as e:
            problematicLines.append(line + " --> variable not present in this context.")
        except Exception as e:
            problematicLines.append(line + " --> " + str(e))

    if problematicLines:
        return PyworkflowException('Following lines in %s seems to be problematic. '
                                   'Please review its format or content.\n%s' % (pw.Config.SCIPION_HOSTS, "\n".join(problematicLines)),
                                   url=pw.DOCSITEURLS.HOST_CONFIG)

def _submit(hostConfig, submitDict, cwd=None, env=None):
    """ Submit a protocol to a queue system. Return its job id.
    """
    # Create first the submission script to be launched
    # formatting using the template
    template = hostConfig.getSubmitTemplate()

    try:
        template = template % submitDict
    except Exception as e:
        # Capture parsing errors
        exception = analyzeFormattingTypeError(template, submitDict)

        if exception:
            raise exception
        else:
            # If there is no exception, then raise actual one
            raise e

    # FIXME: CREATE THE PATH FIRST
    scripPath = submitDict['JOB_SCRIPT']
    f = open(scripPath, 'w')
    # Ensure the path exists
    makeFilePath(scripPath)
    # Add some line ends because in some clusters it fails
    # to submit jobs if the submit script does not have end of line
    f.write(template + '\n\n')
    f.close()
    # This should format the command using a template like: 
    # "qsub %(JOB_SCRIPT)s"
    command = hostConfig.getSubmitCommand() % submitDict
    gcmd = greenStr(command)
    print("** Submitting to queue: '%s'" % gcmd)

    p = Popen(command, shell=True, stdout=PIPE, cwd=cwd, env=env)
    out = p.communicate()[0]
    # Try to parse the result of qsub, searching for a number (jobId)
    # Review this, seems to exclusive to torque batch system
    s = re.search('(\d+)', str(out))
    if p.returncode == 0 and s:
        job = int(s.group(0))
        print("Launched job with id %s" % job)
        return job
    else:
        print("Couldn't submit to queue for reason: %s " % redStr(out.decode()))
        return UNKNOWN_JOBID


def _run(command, wait, stdin=None, stdout=None, stderr=None):
    """ Execute a command in a subprocess and return the pid. """
    gcmd = greenStr(command)
    print("** Running command: '%s'" % gcmd)
    p = Popen(command, shell=True, stdout=stdout, stderr=stderr)
    jobId = p.pid
    if wait:
        p.wait()

    return jobId


# ******************************************************************
# *                 Function related to STOP
# ******************************************************************


def _stopLocal(protocol):
    if protocol.useQueue() and not protocol.isScheduled():
        jobId = protocol.getJobId()
        host = protocol.getHostConfig()
        cancelCmd = host.getCancelCommand() % {'JOB_ID': jobId}
        _run(cancelCmd, wait=True)
    else:
        process.killWithChilds(protocol.getPid())


def _stopRemote(protocol):
    _runRemote(protocol, 'stop')



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *              Roberto Marabini       (roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from os.path import exists

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET

from .mapper import Mapper


class XmlMapper(Mapper):
    """Mapper for XML"""
    def __init__(self, filename, dictClasses=None, rootName='ALL', **args):
        self.filename = filename        
        if exists(filename):
            self._read()           
        else:
            self._create(rootName, **args)
        Mapper.__init__(self, dictClasses)
        # Objects map (id should be defined)  
        self.objDict = {}      
        # Store some objects during parsing
        # that have some Pointer attributes and need to be fixed later 
        self.pendingPtrDict = {}
        # This dictionary serve to define how to write classes
        # for example if the pair 'Integer': 'attribute' is present
        # all Integer will be store as attributes in the xml
        # possible values are: 
        #   'attribute', 'class_only', 'class_name', 'name_class', 'name_only'
        self.classTags = {} 
        # Counter to provide default objects id's
        self.objCount = 0

    def setClassTag(self, classname, tag):
        self.classTags[classname] = tag
        
    def indent(self, elem, level=0):
        i = "\n" + level*"  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for _elem in elem:
                self.indent(_elem, level+1)
            if not _elem.tail or not _elem.tail.strip():
                _elem.tail = i
        else:
            # TODO: generalize the matching of consecutive items
            if level and (not elem.tail or not elem.tail.strip()):
                for ii in ['t11', 't12', 't13', 't21', 't22', 't23', 't31', 't32', 't33']:
                    if elem.tag == ii:
                        elem.tail = " "
                        return
                elem.tail = i
    
    def _create(self, rootName, **args):
        self.root = ET.Element(rootName)
        if 'header' in args:
            self.root.append(ET.Comment(args.get('header')))
            if 'version' in args:
                self.root.set("version", str(args.get('version')))
    
    def _read(self):
        if self.filename is None:
            raise Exception("XmlMapper.read: filename is None.")
        if not exists(self.filename):
            raise Exception("XmlMapper.read: filename '%s' does not exists.")
        self.tree = ET.parse(self.filename)
        self.root = self.tree.getroot()

    def strId(self, objId):
        """Create an unique id from the objId,
        specially for dictionaries, where all keys
        and values will be converted to string"""
        if type(objId) is dict:
            ordKeys = objId.keys()
            ordKeys.sort()
            sid = ''
            for k in ordKeys:
                sid += '%s:%s ' % (k, objId[k])
            return sid
        return str(objId)
        
    def selectById(self, objId):
        """Retrieve an object give the id"""
        return self.objDict.get(self.strId(objId), None)
    
    def _addObjectToDict(self, obj, objDict):
        """Add object if have id"""
        if obj.hasObjId():
            key = self.strId(obj.getObjId())
        else:
            self.objCount += 1
            key = '%s_%06d' % (obj.getClassName(), self.objCount)
        objDict[key] = obj
        
    def selectAll(self, iterate=False):
        """Select object from storage"""
        for child in self.root:
            obj = self._buildObjectFromClass(child.tag)
            self.fillObject(obj, child)
            self._addObjectToDict(obj, self.objDict)
        # set properly primary keys
        for obj in self.pendingPtrDict.values():
            for key, attr in obj.getAttributesToStore():
                if attr.isPointer(): 
                    ptr = self.selectById(attr._objId)
                    attr._objId = None
                    attr.set(ptr)
        return self.objDict.values()
                   
    def setChildObject(self, obj, childName, childClass=None):
        childObj = getattr(obj, childName, None)
        if childObj is None:
            if childClass is None:
                raise Exception('Attribute "%s" was not found in parent object and classname not stored' % childName)
            childObj = self._buildObjectFromClass(childClass)
            setattr(obj, childName, childObj)
        return childObj
                  
    def fillObject(self, obj, objElem):
        # Set attributes first
        if not obj.isPointer():
            for k, v in objElem.attrib.items():
                if k not in ['id', 'classname', 'attrname']:
                    childObj = self.setChildObject(obj, k)
                    childObj.set(v)
        # Set tree childs attributes
        for child in objElem:
            if 'classname' in child.attrib:
                childName = child.tag
                childClass = child.attrib['classname']
            elif 'attrname' in child.attrib:
                childName = child.attrib['attrname']
                childClass = child.tag
            else:
                tagKey = '%s.%s' % (obj.getClassName(), child.tag)
                
                tag = self.classTags.get(tagKey, None)
                if tag == 'name_only' or child.tag not in self.dictClasses:
                    childName = child.tag
                    childClass = None
                else:
                    childName = ''
                    childClass = child.tag
                
            childObj = self.setChildObject(obj, childName, childClass)
            
            if child.text and len(child.text.strip()):
                childObj.set(child.text.strip())
            else:
                # does have attributes?
                attributes = child.attrib
                if attributes and childObj.isPointer():
                    childObj._objId = attributes
                    # save obj in auxiliary file
                    self._addObjectToDict(obj, self.pendingPtrDict)
                self.fillObject(childObj, child)
        
    def commit(self):
        if self.filename is None:
            raise Exception("XmlMapper.commit: filename is None")
        self.indent(self.root)
        tree = ET.ElementTree(self.root)
        tree.write(self.filename,
                   xml_declaration=True,
                   encoding='utf-8')

    def setObjectId(self, objElem, objId):
        # Set attributes of this object element
        if objId:
            if type(objId) is dict:
                for k, v in objId.items():
                    if v:
                        objElem.set(k, str(v))
            else:        
                objElem.set('id', str(objId))
        
    def insert(self, obj):
        """Insert a new object into the system"""
        objElem = self.addSubElement(self.root, obj.getClassName(), obj._objValue) 
        self.setObjectId(objElem, obj.getObjId())
        # Insert object childs
        self.insertObjectWithChilds(obj, objElem)
        
    def addSubElement(self, parentElem, name, value=None):
        childElem = ET.SubElement(parentElem, name)
        if value is not None:
            childElem.text = str(value)
        return childElem
            
    def insertObjectWithChilds(self, obj, parentElem):
        
        for key, attr in obj.getAttributesToStore():            
            if attr.hasValue():
                attrClass = attr.getClassName()
                objClass = obj.getClassName()
                allKey = '%s.ALL' % objClass  # First try with .ALL
                tag = self.classTags.get(allKey, '')
                allKey = 'ALL.%s' % attrClass 
                tag = self.classTags.get(allKey, tag)
                allKey = 'ALL.%s' % key  # also with ALL.attribute
                tag = self.classTags.get(allKey, tag)
                classKey = '%s.%s' % (objClass, attrClass)  # Second, with .childClass
                tag = self.classTags.get(classKey, tag)
                attrKey = '%s.%s' % (objClass, key)  # Finally with .attribute
                tag = self.classTags.get(attrKey, tag)
                
                if tag == 'attribute':
                    parentElem.set(key, str(attr))
                else:
                    if attr.isPointer():
                        childElem = self.addSubElement(parentElem, key)
                        self.setObjectId(childElem, attr.get().getObjId())
                    else:
                        # Select if we want to use the classname or elem name
                        # as element tagname
                        if tag.startswith('class'):
                            name = attrClass
                        else:
                            name = key
                        childElem = self.addSubElement(parentElem, name, attr._objValue)
                        if tag.endswith('class'):
                            childElem.set('classname', attrClass)
                        elif tag.endswith('name'):
                            childElem.set('attrname', key)
                        
                        self.insertObjectWithChilds(attr, childElem)
            
    def updateFrom(self, obj):
        """Update object data with storage info"""
        pass
    
    def updateTo(self, obj):
        """Update storage with object info"""
        pass



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *              Jose Gutierrez (jose.gutierrez@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Module to handling Dialogs
some code was taken from tkSimpleDialog
"""
import tkinter as tk
import traceback
# from tkinter.colorchooser import askcolor as _askColor
from tkcolorpicker import askcolor as _askColor

from pyworkflow.exceptions import PyworkflowException
from pyworkflow.utils import Message, Icon
from . import gui
from .tree import BoundTree
from .text import Text, TaggedText

# Possible result values for a Dialog
RESULT_YES = 0
RESULT_NO = 1
RESULT_CANCEL = 2
RESULT_RUN_SINGLE = 3
RESULT_RUN_ALL = 4
RESULT_CLOSE = 5


class Dialog(tk.Toplevel):
    _images = {}  # Images cache
    """Implementation of our own dialog to display messages
    It will have by default a three buttons: YES, NO and CANCEL
    Subclasses can rename the labels of the buttons like: OK, CLOSE or others
    The buttons(and theirs order) can be changed.
    An image name can be passed to display left to the message.
    """

    def __init__(self, parent, title, **kwargs):
        """Initialize a dialog.
        Arguments:
            parent -- a parent window (the application window)
            title -- the dialog title
        **args accepts:
            buttons -- list of buttons tuples containing which buttons to display
        """

        if parent is None:
            parent = tk.Tk()
            parent.withdraw()
            gui.setCommonFonts()

        tk.Toplevel.__init__(self, parent)

        self.withdraw()  # remain invisible for now
        # If the master is not viewable, don't
        # make the child transient, or else it
        # would be opened withdrawn
        if parent.winfo_viewable():
            self.transient(parent)

        if title:
            self.title(title)

        self.parent = parent
        # Default to CANCEL so if window is "Closed" behaves the same.
        self.result = RESULT_CANCEL
        self.initial_focus = None

        bodyFrame = tk.Frame(self)
        # Call subclass method body to create that region
        self.body(bodyFrame)
        bodyFrame.grid(row=0, column=0, sticky='news',
                       padx=5, pady=5)

        self.icons = kwargs.get('icons',
                                {RESULT_YES: Icon.BUTTON_SELECT,
                                 RESULT_NO: Icon.BUTTON_CLOSE,
                                 RESULT_CANCEL: Icon.BUTTON_CANCEL,
                                 RESULT_CLOSE: Icon.BUTTON_CLOSE})

        self.buttons = kwargs.get('buttons', [('OK', RESULT_YES),
                                              ('Cancel', RESULT_CANCEL)])
        self.defaultButton = kwargs.get('default', 'OK')
        btnFrame = tk.Frame(self)
        # Create buttons 
        self.buttonbox(btnFrame)
        btnFrame.grid(row=1, column=0, sticky='sew',
                      padx=5, pady=(0, 5))

        gui.configureWeigths(self)

        if self.initial_focus is None:
            self.initial_focus = self

        self.protocol("WM_DELETE_WINDOW", self.cancel)

        if self.parent is not None:
            position = kwargs.get('position', (parent.winfo_rootx() + 50,
                                               parent.winfo_rooty() + 50))
            self.geometry("+%d+%d" % position)

        self.deiconify()  # become visible now
        self.initial_focus.focus_set()
        # Pablo: I've commented this when migrating to python3 since I was getting and exception:
        # window ".139897767953072.139897384058440" was deleted before its visibility changed
        # wait for window to appear on screen before calling grab_set
        self.wait_visibility()
        self.grab_set()
        self.wait_window(self)

    def destroy(self):
        """Destroy the window"""
        self.initial_focus = None
        tk.Toplevel.destroy(self)

    #
    # construction hooks

    def body(self, master):
        """create dialog body.
        return widget that should have initial focus.
        This method should be overridden, and is called
        by the __init__ method.
        """
        pass

    def _createButton(self, frame, text, result):
        icon = None
        if result in self.icons.keys():
            icon = self.getImage(self.icons[result])
        return tk.Button(frame, text=text, image=icon,
                         compound=tk.LEFT,
                         command=lambda: self._handleResult(result))

    def buttonbox(self, btnFrame):
        frame = tk.Frame(btnFrame)
        btnFrame.columnconfigure(0, weight=1)
        frame.grid(row=0, column=0)
        col = 0
        for btnLabel, btnResult in self.buttons:
            btn = self._createButton(frame, btnLabel, btnResult)
            btn.grid(row=0, column=col, padx=5, pady=5)
            if (btnLabel == self.defaultButton and
                    self.initial_focus is None):
                self.initial_focus = btn
            col += 1
        self.bind("<Return>", self._handleReturn)
        self.bind("<KP_Enter>", self._handleReturn)
        self.bind("<Escape>", lambda e: self._handleResult(RESULT_CANCEL))

    def _handleResult(self, resultValue):
        """This method will be called when any button is pressed.
        It will set the resultValue associated with the button
        and close the Dialog"""
        self.result = resultValue
        noCancel = self.result != RESULT_CANCEL and self.result != RESULT_CLOSE

        if noCancel and not self.validate():
            self.initial_focus.focus_set()  # put focus back
            return

        self.withdraw()
        self.update_idletasks()

        try:
            if noCancel:
                self.apply()
        finally:
            self.cancel()

    def _handleReturn(self, e=None):
        """Handle press return key"""
        # Check which of the buttons is the default
        for button, result in self.buttons:
            if self.defaultButton == button:
                self._handleResult(result)

    def cancel(self, event=None):
        # put focus back to the parent window
        if self.parent is not None:
            self.parent.focus_set()
        self.destroy()

    #
    # command hooks

    def validate(self):
        """validate the data
        This method is called automatically to validate the data before the
        dialog is destroyed. By default, it always validates OK.
        """
        return 1  # override

    def apply(self):
        """process the data
        This method is called automatically to process the data, *after*
        the dialog is destroyed. By default, it does nothing.
        """
        pass  # override

    def getImage(self, imgName):
        """A shortcut to get an image from its name"""
        return gui.getImage(imgName, self._images)

    def getResult(self):
        return self.result

    def resultYes(self):
        return self.result == RESULT_YES

    def resultNo(self):
        return self.result == RESULT_NO

    def resultCancel(self):
        return self.result == RESULT_CANCEL


def fillMessageText(text, message):
    # Insert lines of text
    if isinstance(message, list):
        lines = message
    else:
        lines = message.splitlines()
    text.setReadOnly(False)
    text.clear()
    w = 0
    for l in lines:
        w = max(w, len(l))
        text.addLine(l)
    w = min(w + 5, 80)
    h = min(len(lines) + 3, 30)
    text.config(height=h, width=w)
    text.addNewline()
    text.setReadOnly(True)


def createMessageBody(bodyFrame, message, image,
                      frameBg='white',
                      textBg='white',
                      textPad=5):
    """ Create a Text containing the message.
    Params:
        bodyFrame: tk.Frame to be filled.
        msg: a str or list with the lines.
    """
    bodyFrame.config(bg=frameBg, bd=0)
    text = TaggedText(bodyFrame, bg=textBg, bd=0, highlightthickness=0)
    # Insert image
    if image:
        label = tk.Label(bodyFrame, image=image, bg=textBg, bd=0)
        label.grid(row=0, column=0, sticky='nw')

    text.frame.grid(row=0, column=1, sticky='news',
                    padx=textPad, pady=textPad)
    fillMessageText(text, message)
    bodyFrame.rowconfigure(0, weight=1)
    bodyFrame.columnconfigure(1, weight=1)

    return text


class MessageDialog(Dialog):
    """Dialog subclasses to show message info, questions or errors.
    It can display an icon with the message"""

    def __init__(self, parent, title, msg, iconPath, **args):
        self.msg = msg
        self.iconPath = iconPath
        if 'buttons' not in args:
            args['buttons'] = [('OK', RESULT_YES)]
            args['default'] = 'OK'
        Dialog.__init__(self, parent, title, **args)

    def body(self, bodyFrame):
        self.image = gui.getImage(self.iconPath)
        createMessageBody(bodyFrame, self.msg, self.image)


class ExceptionDialog(MessageDialog):
    def __init__(self, *args, **kwargs):
        self._exception = None if "exception" not in kwargs else kwargs['exception']
        super().__init__(*args, **kwargs)

    def body(self, bodyFrame):
        super().body(bodyFrame)

        def addTraceback(event):
            detailsText = TaggedText(bodyFrame, bg='white', bd=0, highlightthickness=0)
            traceStr = traceback.format_exc()
            fillMessageText(detailsText, traceStr)
            detailsText.frame.grid(row=row+1, column=0, columnspan=2, sticky='news', padx=5, pady=5)
            event.widget.grid_forget()

        row = 1
        if self._exception:

            if isinstance(self._exception, PyworkflowException):
                helpUrl = self._exception.getUrl()
                labelUrl = TaggedText(bodyFrame, bg='white', bd=0, highlightthickness=0)
                fillMessageText(labelUrl,"Please go here for more details: %s" % helpUrl)
                labelUrl.grid(row=row, column=0, columnspan=2, sticky='news')
                row += 1

            label = tk.Label(bodyFrame, text="Show details...", bg='white', bd=0)
            label.grid(row=row, column=0, columnspan=2, sticky='news')
            label.bind("<Button-1>", addTraceback)


class YesNoDialog(MessageDialog):
    """Ask a question with YES/NO answer"""

    def __init__(self, master, title, msg, **kwargs):
        buttonList = [('Yes', RESULT_YES), ('No', RESULT_NO)]

        if kwargs.get('showCancel', False):
            buttonList.append(('Cancel', RESULT_CANCEL))

        MessageDialog.__init__(self, master, title, msg,
                               'fa-exclamation-triangle_alert.gif', default='No',
                               buttons=buttonList)


class GenericDialog(Dialog):
    """
    Create a dialog with many buttons
    Arguments:
            parent -- a parent window (the application window)
            title -- the dialog title
            msg  -- message to display into the dialog
            iconPath -- path of the image to show into the dialog

        **args accepts:
            buttons -- list of buttons tuples containing which buttons to display and theirs values
            icons -- list of icons for all buttons
            default -- button default

            Example:
                buttons=[('Single', RESULT_RUN_SINGLE),
                         ('All', RESULT_RUN_ALL),
                         ('Cancel', RESULT_CANCEL)],
                default='Cancel',
                icons={RESULT_CANCEL: Icon.BUTTON_CANCEL,
                       RESULT_RUN_SINGLE: Icon.BUTTON_SELECT,
                       RESULT_RUN_ALL: Icon.ACTION_EXECUTE})
    """

    def __init__(self, master, title, msg, iconPath, **kwargs):
        self.msg = msg
        self.iconPath = iconPath
        Dialog.__init__(self, master, title, **kwargs)

    def body(self, bodyFrame):
        self.image = gui.getImage(self.iconPath)
        createMessageBody(bodyFrame, self.msg, self.image)


class EntryDialog(Dialog):
    """Dialog to ask some entry"""

    def __init__(self, parent, title, entryLabel, entryWidth=20,
                 defaultValue='', headerLabel=None):
        self.entryLabel = entryLabel
        self.entryWidth = entryWidth
        self.headerLabel = headerLabel
        self.tkvalue = tk.StringVar()
        self.tkvalue.set(defaultValue)
        self.value = None
        Dialog.__init__(self, parent, title)

    def body(self, bodyFrame):
        bodyFrame.config(bg='white')
        frame = tk.Frame(bodyFrame, bg='white')
        frame.grid(row=0, column=0, padx=20, pady=20)
        row = 0
        if self.headerLabel:
            label = tk.Label(bodyFrame, text=self.headerLabel, bg='white', bd=0)
            label.grid(row=row, column=0, columnspan=2, sticky='nw', padx=(15, 10), pady=15)
            row += 1
        label = tk.Label(bodyFrame, text=self.entryLabel, bg='white', bd=0)
        label.grid(row=row, column=0, sticky='nw', padx=(15, 10), pady=15)
        self.entry = tk.Entry(bodyFrame, bg=gui.cfgEntryBgColor,
                              width=self.entryWidth, textvariable=self.tkvalue)
        self.entry.grid(row=row, column=1, sticky='new', padx=(0, 15), pady=15)
        self.initial_focus = self.entry

    def apply(self):
        self.value = self.entry.get()

    def validate(self):
        if len(self.entry.get().strip()) == 0:
            showError("Validation error", "Value is empty", self)
            return False
        return True


class EditObjectDialog(Dialog):
    """Dialog to edit some text"""

    def __init__(self, parent, title, obj, mapper, **kwargs):
        self.obj = obj
        self.mapper = mapper

        self.textWidth = 5
        self.textHeight = 1
        self.labelText = kwargs.get('labelText', Message.TITLE_LABEL)
        self.valueText = self.obj.getObjLabel()

        self.commentLabel = Message.TITLE_COMMENT
        self.commentWidth = 50
        self.commentHeight = 15
        self.valueComment = self.obj.getObjComment()

        Dialog.__init__(self, parent, title, **kwargs)

    def body(self, bodyFrame):
        bodyFrame.config(bg='white')
        frame = tk.Frame(bodyFrame, bg='white')
        frame.grid(row=0, column=0, padx=20, pady=20)

        # Label
        label_text = tk.Label(bodyFrame, text=self.labelText, bg='white', bd=0)
        label_text.grid(row=0, column=0, sticky='nw', padx=(15, 10), pady=15)
        # Label box
        var = tk.StringVar()
        var.set(self.valueText)
        self.textLabel = tk.Entry(bodyFrame, width=self.textWidth, textvariable=var, font=gui.getDefaultFont())
        self.textLabel.grid(row=0, column=1, sticky='news', padx=5, pady=5)

        # Comment
        label_comment = tk.Label(bodyFrame, text=self.commentLabel, bg='white', bd=0)
        label_comment.grid(row=1, column=0, sticky='nw', padx=(15, 10), pady=15)
        # Comment box
        self.textComment = Text(bodyFrame, height=self.commentHeight,
                                width=self.commentWidth)
        self.textComment.setReadOnly(False)
        self.textComment.setText(self.valueComment)
        self.textComment.grid(row=1, column=1, sticky='news', padx=5, pady=5)
        self.initial_focus = self.textLabel

    def getLabel(self):
        return self.textLabel.get()

    def getComment(self):
        return self.textComment.getText()

    def apply(self):
        self.obj.setObjLabel(self.getLabel())
        self.obj.setObjComment(self.getComment())

        if self.obj.hasObjId():
            self.mapper.store(self.obj)
            self.mapper.commit()

    def buttonbox(self, btnFrame):
        # Cancel the binding of <Return> key
        Dialog.buttonbox(self, btnFrame)
        # self.bind("<Return>", self._noReturn)
        self.unbind("<Return>")

    def _noReturn(self, e):
        pass


""" Functions to display dialogs """


def askYesNo(title, msg, parent):
    d = YesNoDialog(parent, title, msg)
    return d.resultYes()


def askYesNoCancel(title, msg, parent):
    d = YesNoDialog(parent, title, msg, showCancel=True)
    return d.result


def askSingleAllCancel(title, msg, parent):
    d = GenericDialog(parent, title, msg,
                      'fa-exclamation-triangle_alert.gif',
                      buttons=[('Single', RESULT_RUN_SINGLE),
                               ('All', RESULT_RUN_ALL),
                               ('Cancel', RESULT_CANCEL)],
                      default='Single',
                      icons={RESULT_CANCEL: Icon.BUTTON_CANCEL,
                             RESULT_RUN_SINGLE: Icon.BUTTON_SELECT,
                             RESULT_RUN_ALL: Icon.ACTION_EXECUTE})

    return d.result


def showInfo(title, msg, parent):
    MessageDialog(parent, title, msg, 'fa-info-circle_alert.gif')


def showWarning(title, msg, parent):
    MessageDialog(parent, title, msg, 'fa-exclamation-triangle_alert.gif')


def showError(title, msg, parent, exception=None):
    ExceptionDialog(parent, title, msg, 'fa-times-circle_alert.gif', exception=exception)


def askString(title, label, parent, entryWidth=20, defaultValue='', headerLabel=None):
    d = EntryDialog(parent, title, label, entryWidth, defaultValue, headerLabel)
    return d.value


def askColor(parent, defaultColor='black'):
    (rgbcolor, hexcolor) = _askColor(defaultColor, parent=parent)
    return hexcolor


class ListDialog(Dialog):
    """
    Dialog to select an element from a list.
    It is implemented using the Tree widget.
    """

    def __init__(self, parent, title, provider, message=None, **kwargs):
        """ From kwargs:
                message: message tooltip to show when browsing.
                selected: the item that should be selected.
                validateSelectionCallback:
                    a callback function to validate selected items.
                allowSelect: if set to False, the 'Select' button will not
                    be shown.
                allowsEmptySelection: if set to True, it will not validate
                    that at least one element was selected.
        """
        self.values = []
        self.provider = provider
        self.message = message
        self.validateSelectionCallback = kwargs.get('validateSelectionCallback',
                                                    None)
        self._selectmode = kwargs.get('selectmode', 'extended')
        self._selectOnDoubleClick = kwargs.get('selectOnDoubleClick', False)
        self._allowsEmptySelection = kwargs.get('allowsEmptySelection', False)

        buttons = []
        if kwargs.get('allowSelect', True):
            buttons.append(('Select', RESULT_YES))
        if kwargs.get('cancelButton', False):
            buttons.append(('Close', RESULT_CLOSE))
        else:
            buttons.append(('Cancel', RESULT_CANCEL))

        Dialog.__init__(self, parent, title, buttons=buttons, **kwargs)

    def body(self, bodyFrame):
        bodyFrame.config()
        gui.configureWeigths(bodyFrame)
        dialogFrame = tk.Frame(bodyFrame)
        dialogFrame.grid(row=0, column=0, sticky='news', padx=5, pady=5)
        dialogFrame.config()
        gui.configureWeigths(dialogFrame, row=1)
        self._createFilterBox(dialogFrame)
        self._createTree(dialogFrame)
        if self.message:
            label = tk.Label(bodyFrame, text=self.message, compound=tk.LEFT,
                             image=self.getImage(Icon.LIGHTBULB))
            label.grid(row=2, column=0, sticky='nw', padx=5, pady=5)
        self.initial_focus = self.tree

    def _createTree(self, parent):
        self.tree = BoundTree(parent, self.provider, selectmode=self._selectmode)
        if self._selectOnDoubleClick:
            self.tree.itemDoubleClick = lambda obj: self._handleResult(RESULT_YES)
        self.tree.grid(row=1, column=0)

    def _createFilterBox(self, content):
        """ Create the Frame with Filter widgets """

        def _onSearch(e=None):

            def comparison():
                pattern = self._searchVar.get().lower()
                return [w[0] for w in self.lista.items()
                        if pattern in self.lista.get(w[0]).lower()]

            self.tree.update()
            self.lista = {}

            for item in self.tree.get_children():

                itemStr = self.tree.item(item)['text']
                for value in self.tree.item(item)['values']:
                    if isinstance(value, int):
                        itemStr = itemStr + ' ' + str(value)
                    else:
                        itemStr = itemStr + ' ' + value

                self.lista[item] = itemStr

            if self._searchVar.get() != '':
                matchs = comparison()
                if matchs:
                    for item in self.tree.get_children():
                        if item not in matchs:
                            self.tree.delete(item)
                else:
                    self.tree.delete(*self.tree.get_children())

        self.searchBoxframe = tk.Frame(content)
        label = tk.Label(self.searchBoxframe, text="Filter")
        label.grid(row=0, column=0, sticky='nw')
        self._searchVar = tk.StringVar(value='')
        self.entry = tk.Entry(self.searchBoxframe, bg='white',
                              textvariable=self._searchVar, width=40,
                              font=gui.getDefaultFont())

        self.entry.bind('<KeyRelease>', _onSearch)
        self.entry.focus_set()
        self.entry.grid(row=0, column=1, sticky='news')
        self.searchBoxframe.grid(row=0, column=0, sticky='news', padx=5,
                                 pady=(10, 5))

    def apply(self):
        self.values = self.tree.getSelectedObjects()

    def validate(self):
        self.apply()  # load self.values with selected items
        err = ''

        if self.values:
            if self.validateSelectionCallback:
                err = self.validateSelectionCallback(self.values)
        else:
            if not self._allowsEmptySelection:
                err = "Please select an element"

        if err:
            showError("Validation error", err, self)
            return False

        return True


class ToolbarButton:
    """
    Store information about the buttons that will be added to the toolbar.
    """

    def __init__(self, text, command, icon=None, tooltip=None):
        self.text = text
        self.command = command
        self.icon = icon
        self.tooltip = tooltip


class ToolbarListDialog(ListDialog):
    """
    This class extend from ListDialog to allow an
    extra toolbar to handle operations over the elements
    in the list (e.g. Edit, New, Delete).
    """

    def __init__(self, parent, title, provider,
                 message=None, toolbarButtons=None, **kwargs):
        """ From kwargs:
                message: message tooltip to show when browsing.
                selected: the item that should be selected.
                validateSelectionCallback:
                    a callback function to validate selected items.
                allowSelect: if set to False, the 'Select' button will not
                    be shown.
        """
        self.toolbarButtons = toolbarButtons
        self._itemDoubleClick = kwargs.get('itemDoubleClick', None)
        ListDialog.__init__(self, parent, title, provider, message, **kwargs)

    def body(self, bodyFrame):
        bodyFrame.config(bg='white')
        gui.configureWeigths(bodyFrame, 1, 0)

        # Add an extra frame to insert the Toolbar
        # and another one for the ListDialog's body
        self.toolbarFrame = tk.Frame(bodyFrame, bg='white')
        self.toolbarFrame.grid(row=0, column=0, sticky='new')
        if self.toolbarButtons:
            for i, b in enumerate(self.toolbarButtons):
                self._addButton(b, i)

        subBody = tk.Frame(bodyFrame)
        subBody.grid(row=1, column=0, sticky='news', padx=5, pady=5)
        ListDialog.body(self, subBody)
        if self._itemDoubleClick:
            self.tree.itemDoubleClick = self._itemDoubleClick

    def _addButton(self, button, col):
        btn = tk.Label(self.toolbarFrame, text=button.text,
                       image=self.getImage(button.icon),
                       compound=tk.LEFT, cursor='hand2', bg='white')
        btn.grid(row=0, column=col, sticky='nw', padx=(5, 0), pady=(5, 0))
        btn.bind('<Button-1>', button.command)


class FlashMessage:
    def __init__(self, master, msg, delay=5, relief='solid', func=None):
        self.root = tk.Toplevel(master=master)
        # hides until know geometry
        self.root.withdraw()
        self.root.wm_overrideredirect(1)
        tk.Label(self.root, text="   %s   " % msg,
                 bd=1, bg='DodgerBlue4', fg='white').pack()
        gui.centerWindows(self.root, refWindows=master)
        self.root.deiconify()
        self.root.grab_set()
        self.msg = msg

        if func:
            self.root.update_idletasks()
            self.root.after(10, self.process, func)
        else:
            self.root.after(int(delay * 1000), self.close)
        self.root.wait_window(self.root)

    def process(self, func):
        func()
        self.root.destroy()

    def close(self):
        self.root.destroy()


class FloatingMessage:
    def __init__(self, master, msg, xPos=750, yPos=80, textWidth=280,
                 font='Helvetica', size=12, bd=1, bg='#6E6E6E', fg='white'):
        self.floatingMessage = tk.Label(master, text="   %s   " % msg,
                                        bd=bd, bg=bg, fg=fg)
        self.floatingMessage.place(x=xPos, y=yPos, width=textWidth)
        self.floatingMessage.config(font=(font, size))

    def show(self):
        self.floatingMessage.update_idletasks()

    def close(self):
        self.floatingMessage.destroy()


class FileBrowseDialog(Dialog):
    """Dialog to select files from the filesystem."""

    def __init__(self, parent, title, provider, message=None, **args):
        """ From args:
                message: message tooltip to show when browsing.
                selected: the item that should be selected.
        """
        self.value = None
        self.provider = provider
        self.message = args.get('message', None)
        Dialog.__init__(self, parent, title,
                        buttons=[('Select', RESULT_YES), ('Cancel', RESULT_CANCEL)])

    def body(self, bodyFrame):
        bodyFrame.config(bg='white')
        gui.configureWeigths(bodyFrame)
        self._createTree(bodyFrame)
        if self.message:
            label = tk.Label(bodyFrame, text=self.message, bg='white',
                             image=self.getImage('fa-lightbulb-o.gif'), compound=tk.LEFT)
            label.grid(row=1, column=0, sticky='nw', padx=5, pady=5)
        self.initial_focus = self.tree

    def _createTree(self, parent):
        self.tree = BoundTree(parent, self.provider)

    def apply(self):
        index = self.tree.index(self.tree.getFirst())
        self.value = self.tree._objects[index]

    def validate(self):
        if self.tree.getFirst() is None:
            showError("Validation error", "Please select an element", self)
            return False
        return True



# coding: latin-1
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""

@article{delaRosaTrevin201693,
title = "Scipion: A software framework toward integration, reproducibility and validation in 3D electron microscopy ",
journal = "Journal of Structural Biology",
volume = "195",
number = "1",
pages = "93 - 99",
year = "2016",
note = "",
issn = "1047-8477",
doi = "http://doi.org/10.1016/j.jsb.2016.04.010",
url = "http://www.sciencedirect.com/science/article/pii/S104784771630079X",
author = "J.M. de la Rosa-Trevín and A. Quintana and L. del Cano and A. Zaldívar and I. Foche and J. Gutiérrez and
J. Gómez-Blanco and J. Burguet-Castell and J. Cuenca-Alba and V. Abrishami and J. Vargas and J. Otón and G. Sharov
and J.L. Vilas and J. Navas and P. Conesa and M. Kazemi and R. Marabini and C.O.S. Sorzano and J.M. Carazo",
keywords = "Electron microscopy",
keywords = "Single particle analysis",
keywords = "Image processing",
keywords = "Software package",
keywords = "Workflows",
keywords = "Reproducibility "
}

"""



#!/usr/bin/env python

import sys
import os

from pyworkflow.project import Manager
import pyworkflow.utils as pwutils


def usage(error):
    print("""
    ERROR: %s
    
    Usage: python -m pyworkflow.project.script.fix_links PROJECT SEARCH_DIR
        PROJECT: provide the project name to fix broken links in the imports.
        SEARCH_DIR: provide a directory where to look for the files.
        and fix the links.    
    """ % error)
    os._exit(1)

def main():
    if len(sys.argv) != 3:
        usage("Incorrect number of input parameters")

    projName = sys.argv[1]
    searchDir = os.path.abspath(sys.argv[2])

    # Create a new project
    manager = Manager()

    if not manager.hasProject(projName):
        usage("Nonexistent project: %s" % pwutils.red(projName))

    if not os.path.exists(searchDir):
        usage("Nonexistent SEARCH_DIR: %s" % pwutils.red(searchDir))

    project = manager.loadProject(projName)

    project.fixLinks(searchDir)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import logging
import unittest

from pyworkflow import Config
from pyworkflow.utils import (getLineInFile,
                              LoggingConfigurator, restoreStdoutAndErr)
from pyworkflow.tests import BaseTest, setupTestOutput
from pyworkflow.utils.process import runJob


class TestLogs(BaseTest):
    
    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)        

    def testSimpleFileLog(self):

        # Default generic configuration
        Config.SCIPION_LOG = self.getOutputPath("general.log")
        genLogFn = Config.SCIPION_LOG
        print("General log file at %s" % genLogFn)
        LoggingConfigurator.setupLogging()
        log1 = logging.getLogger('pyworkflow.test.log.test_scipion_log')

        def testMessage(message, msg_callback, file, shouldExist):

            if msg_callback:
                msg_callback(message)

            self.assertEqual(shouldExist, bool(getLineInFile(message, file)))

        testMessage("INFO to GEN", log1.info, genLogFn, True)
        testMessage("DEBUG missing in GEN", log1.debug, genLogFn, False)
        testMessage("WARNING in GEN", log1.warning, genLogFn, True)
        testMessage("ERROR in GEN", log1.error, genLogFn, True)

        # Protocol run logging configuration (this is propagating the messages,
        # so messages end un in general log too). This is to allow custom configurations to receive running protocol messages)
        logFn = self.getOutputPath('stdout.log')
        logErrFn = self.getOutputPath('stdErr.log')
        log2 = LoggingConfigurator.setUpProtocolRunLogging(logFn, logErrFn)

        fileInfoTest = 'INFO to FILE'
        testMessage(fileInfoTest, log2.info, logFn, True)
        testMessage(fileInfoTest, None, genLogFn, False)

        fileDebugMsg = "DEBUG does not reach FILE nor GEN"
        testMessage(fileDebugMsg, log2.debug, logFn, False)
        testMessage(fileDebugMsg, None, genLogFn, False)

        fileWarningTest = 'WARNING to FILE and not GEN'
        testMessage(fileWarningTest, log2.warning, logFn, True)
        testMessage(fileWarningTest, None, genLogFn, False)

        fileErrorTest = 'ERROR to FILE and not GEN'
        testMessage(fileErrorTest, log2.error, logFn, True)
        testMessage(fileErrorTest, None, genLogFn, False)

        # Test print goes to the log file (stdout is captured)
        printStdOut = "Print ends up in stdout FILE"
        print(printStdOut, flush=True)
        testMessage(printStdOut,None, logFn, True)

        subprocessOut = "subprocess output in stdout FILE"
        runJob(None,"echo", subprocessOut)
        testMessage(subprocessOut,None, logFn, True)

        restoreStdoutAndErr()

if __name__ == '__main__':
    unittest.main()



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es) [1]
# *              Jose Gutierrez (jose.gutierrez@cnb.csic.es) [2]
# *
# * [1] SciLifeLab, Stockholm University
# * [2] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
In this module a simple ObjectBrowser is implemented.
This class can be subclasses to extend its functionality.
A concrete use of ObjectBrowser is FileBrowser, where the
elements to inspect and preview are files.
"""
import os.path
import stat
import tkinter as tk
import time

import pyworkflow.utils as pwutils
from . import gui
from .tree import BoundTree, TreeProvider
from .text import TaggedText, openTextFileEditor
from .widgets import Button, HotButton
from .. import Config

PARENT_FOLDER = ".."


class ObjectBrowser(tk.Frame):
    """ This class will implement a simple object browser.
    Basically, it will display a list of elements at the left
    panel and can display a preview and description on the
    right panel for the selected element.
    An ObjectView will be used to grab information for
    each element such as: icon, preview and description.
    A TreeProvider will be used to populate the list (Tree).
    """

    def __init__(self, parent, treeProvider,
                 showPreview=True, showPreviewTop=True,
                 **args):
        tk.Frame.__init__(self, parent, **args)
        self.treeProvider = treeProvider
        self._lastSelected = None
        gui.configureWeigths(self)
        self.showPreviewTop = showPreviewTop
        # The main layout will be two panes, 
        # At the left containing the elements list
        # and the right containing the preview and description
        p = tk.PanedWindow(self, orient=tk.HORIZONTAL)
        p.grid(row=0, column=0, sticky='news')

        leftPanel = tk.Frame(p)
        self._fillLeftPanel(leftPanel)
        p.add(leftPanel, padx=5, pady=5)
        p.paneconfig(leftPanel, minsize=300)

        if showPreview:
            rightPanel = tk.Frame(p)
            self._fillRightPanel(rightPanel)
            p.add(rightPanel, padx=5, pady=5)
            p.paneconfig(rightPanel, minsize=200)

            # Register a callback when the item is clicked
            self.tree.itemClick = self._itemClicked

    def _fillLeftPanel(self, frame):
        gui.configureWeigths(frame)
        self.tree = BoundTree(frame, self.treeProvider)
        self.tree.grid(row=0, column=0, sticky='news')
        self.itemConfig = self.tree.itemConfig
        self.getImage = self.tree.getImage

    def _fillRightPanel(self, frame):
        frame.columnconfigure(0, weight=1)

        if self.showPreviewTop:
            top = tk.Frame(frame)
            top.grid(row=0, column=0, sticky='news')
            frame.rowconfigure(0, weight=3)
            gui.configureWeigths(top)
            top.rowconfigure(0, minsize=200)
            self._fillRightTop(top)

        bottom = tk.Frame(frame)
        bottom.grid(row=1, column=0, sticky='news')
        frame.rowconfigure(1, weight=1)
        gui.configureWeigths(bottom)
        bottom.rowconfigure(1, weight=1)
        self._fillRightBottom(bottom)

    def _fillRightTop(self, top):
        self.noImage = self.getImage('no-image128.gif')
        self.label = tk.Label(top, image=self.noImage)
        self.label.grid(row=0, column=0, sticky='news')

    def _fillRightBottom(self, bottom):
        self.text = TaggedText(bottom, width=40, height=15, bg='white',
                               takefocus=0)
        self.text.grid(row=0, column=0, sticky='news')

    def _itemClicked(self, obj):
        self._lastSelected = obj
        img, desc = self.treeProvider.getObjectPreview(obj)
        # Update image preview
        if self.showPreviewTop:
            if isinstance(img, str):
                img = self.getImage(img)
            if img is None:
                img = self.noImage
            self.label.config(image=img)
        # Update text preview
        self.text.setReadOnly(False)
        self.text.clear()
        if desc is not None:
            self.text.addText(desc)
        self.text.setReadOnly(True)
        if hasattr(self, 'entryLabel') and not self._lastSelected.isDir():
            self.entryVar.set(self._lastSelected.getFileName())

    def getSelected(self):
        """ Return the selected object. """
        return self._lastSelected


# ------------ Classes and Functions related to File browsing --------------

class FileInfo(object):
    """ This class will store some information about a file.
    It will serve to display files items in the Tree.
    """

    def __init__(self, path, filename):
        self._fullpath = os.path.join(path, filename)
        self._filename = filename
        if os.path.exists(self._fullpath):
            self._stat = os.stat(self._fullpath)
        else:
            self._stat = None

    def isDir(self):
        return stat.S_ISDIR(self._stat.st_mode) if self._stat else False

    def getFileName(self):
        return self._filename

    def getPath(self):
        return self._fullpath

    def getSize(self):
        return self._stat.st_size if self._stat else 0

    def getSizeStr(self):
        """ Return a human readable string of the file size."""
        return pwutils.prettySize(self.getSize()) if self._stat else '0'

    def getDateStr(self):
        return pwutils.dateStr(self.getDate()) if self._stat else '0'

    def getDate(self):
        return self._stat.st_mtime if self._stat else 0

    def isLink(self):
        return os.path.islink(self._fullpath)


class FileHandler(object):
    """ This class will be used to get the icon, preview and info
    from the different types of objects.
    It should be used with FileTreeProvider, where different
    types of handlers can be registered.
    """

    def getFileIcon(self, objFile):
        """ Return the icon name for a given file. """
        if objFile.isDir():
            icon = 'file_folder.gif' if not objFile.isLink() else 'file_folder_link.gif'
        else:
            icon = 'file_generic.gif' if not objFile.isLink() else 'file_generic_link.gif'

        return icon

    def getFilePreview(self, objFile):
        """ Return the preview image and description for the specific object."""
        if objFile.isDir():
            return 'fa-folder-open.gif', None
        return None, None

    def getFileActions(self, objFile):
        """ Return actions that can be done with this object.
        Actions will be displayed in the context menu 
        and the first one will be the default when double-click.
        """
        return []


class TextFileHandler(FileHandler):
    def __init__(self, textIcon):
        FileHandler.__init__(self)
        self._icon = textIcon

    def getFileIcon(self, objFile):
        return self._icon


class SqlFileHandler(FileHandler):
    def getFileIcon(self, objFile):
        return 'file_sqlite.gif'


class FileTreeProvider(TreeProvider):
    """ Populate a tree with files and folders of a given path """

    _FILE_HANDLERS = {}
    _DEFAULT_HANDLER = FileHandler()
    FILE_COLUMN = 'File'
    SIZE_COLUMN = 'Size'

    @classmethod
    def registerFileHandler(cls, fileHandler, *extensions):
        """ Register a FileHandler for a given file extension. 
        Params:
            fileHandler: the FileHandler that will take care of extensions.
            *extensions: the extensions list that will be associated to this
                FileHandler.
        """
        for fileExt in extensions:
            handlersList = cls._FILE_HANDLERS.get(fileExt, [])
            handlersList.append(fileHandler)
            cls._FILE_HANDLERS[fileExt] = handlersList

    def __init__(self, currentDir=None, showHidden=False, onlyFolders=False):
        TreeProvider.__init__(self, sortingColumnName=self.FILE_COLUMN)
        self._currentDir = os.path.abspath(currentDir)
        self._showHidden = showHidden
        self._onlyFolders = onlyFolders
        self.getColumns = lambda: [(self.FILE_COLUMN, 300),
                                   (self.SIZE_COLUMN, 70), ('Time', 150)]

    def getFileHandlers(self, obj):
        filename = obj.getFileName()
        fileExt = pwutils.getExt(filename)
        return self._FILE_HANDLERS.get(fileExt, [self._DEFAULT_HANDLER])

    def getObjectInfo(self, obj):
        filename = obj.getFileName()
        fileHandlers = self.getFileHandlers(obj)
        icon = fileHandlers[0].getFileIcon(obj)

        info = {'key': filename, 'text': filename,
                'values': (obj.getSizeStr(), obj.getDateStr()), 'image': icon
                }

        return info

    def getObjectPreview(self, obj):
        # Look for any preview available
        fileHandlers = self.getFileHandlers(obj)

        for fileHandler in fileHandlers:
            preview = fileHandler.getFilePreview(obj)
            if preview:
                return preview

    def getObjectActions(self, obj):
        fileHandlers = self.getFileHandlers(obj)
        actions = []
        for fileHandler in fileHandlers:
            actions += fileHandler.getFileActions(obj)
        # Always allow the option to open as text
        # specially useful for unknown formats
        fn = obj.getPath()
        actions.append(("Open external Editor",
                        lambda: openTextFileEditor(fn), pwutils.Icon.ACTION_REFERENCES))

        return actions

    def getObjects(self):

        fileInfoList = []
        if not self._currentDir == pwutils.ROOT:
            fileInfoList.append(FileInfo(self._currentDir, PARENT_FOLDER))

        try:
            # This might fail if there is not granted
            files = os.listdir(self._currentDir)

            for f in files:

                fullPath = os.path.join(self._currentDir, f)
                # If f is a file and only need folders
                if self._onlyFolders and not os.path.isdir(fullPath):
                    continue

                # Do not add hidden files if not requested
                if not self._showHidden and f.startswith('.'):
                    continue

                # All ok...add item.
                fileInfoList.append(FileInfo(self._currentDir, f))
        except Exception as e:
            print("Can't list files at " + self._currentDir, e)

        # Sort objects
        fileInfoList.sort(key=self.fileKey, reverse=not self.isSortingAscending())

        return fileInfoList

    def fileKey(self, f):
        sortDict = {self.FILE_COLUMN: 'getFileName',
                    self.SIZE_COLUMN: 'getSize'}
        return getattr(f, sortDict.get(self._sortingColumnName, 'getDate'))()

    def getDir(self):
        return self._currentDir

    def setDir(self, newPath):
        self._currentDir = newPath


# Some constants for the type of selection
# when the file browser is opened

SELECT_NONE = 0  # No selection, just browse files
SELECT_FILE = 1
SELECT_FOLDER = 2
SELECT_PATH = 3  # Can be either file or folder


class FileBrowser(ObjectBrowser):
    """ The FileBrowser is a particular class of ObjectBrowser
    where the "objects" are just files and directories.
    """

    def __init__(self, parent, initialDir='.',
                 selectionType=SELECT_FILE,
                 selectionSingle=True,
                 allowFilter=True,
                 filterFunction=None,
                 previewDim=144,
                 showHidden=False,  # Show hidden files or not?
                 selectButton='Select',  # Change the Select button text
                 entryLabel=None,  # Display an entry for some input
                 entryValue='',  # Display a value in the entry field
                 showInfo=None,  # Used to notify errors or messages
                 shortCuts=None,  # Shortcuts to common locations/paths
                 onlyFolders=False
                 ):
        self.pathVar = tk.StringVar()
        self.pathVar.set(os.path.abspath(initialDir))
        self.pathEntry = None
        self.previousSearch = None
        self.previousSearchTS = None
        self.shortCuts = shortCuts
        self._provider = FileTreeProvider(initialDir, showHidden, onlyFolders)
        self.selectButton = selectButton
        self.entryLabel = entryLabel
        self.entryVar = tk.StringVar()
        self.entryVar.set(entryValue)

        self.showInfo = showInfo or self._showInfo

        ObjectBrowser.__init__(self, parent, self._provider)

        # focuses on the browser in order to allow to move with the keyboard
        self._goDir(os.path.abspath(initialDir))

        if selectionType == SELECT_NONE:
            selectButton = None

        buttonsFrame = tk.Frame(self)
        self._fillButtonsFrame(buttonsFrame)
        buttonsFrame.grid(row=1, column=0)

    def _showInfo(self, msg):
        """ Default way (print to console) to show a message with a given info.
        """
        print(msg)

    def _fillLeftPanel(self, frame):
        """ Redefine this method to include a buttons toolbar and
        also include a filter bar at the bottom of the Tree.
        """
        # Tree with files
        frame.columnconfigure(0, weight=1)

        treeFrame = tk.Frame(frame)
        ObjectBrowser._fillLeftPanel(self, treeFrame)
        # Register the double-click event
        self.tree.itemDoubleClick = self._itemDoubleClick
        # Register keypress event
        self.tree.itemKeyPressed = self._itemKeyPressed

        treeRow = 3
        treeFrame.grid(row=treeRow, column=0, sticky='news')
        # Toolbar frame
        toolbarFrame = tk.Frame(frame)
        self._fillToolbar(toolbarFrame)
        toolbarFrame.grid(row=0, column=0, sticky='new')

        pathFrame = tk.Frame(frame)
        pathLabel = tk.Label(pathFrame, text='Path')
        pathLabel.grid(row=0, column=0, padx=0, pady=3)
        pathEntry = tk.Entry(pathFrame, bg='white', width=65,
                             textvariable=self.pathVar, font=gui.getDefaultFont())
        pathEntry.grid(row=0, column=1, sticky='new', pady=3)
        pathEntry.bind("<Return>", self._onEnterPath)
        pathEntry.bind("<KP_Enter>", self._onEnterPath)
        self.pathEntry = pathEntry
        pathFrame.grid(row=1, column=0, sticky='new')

        # Entry frame, could be used for filter
        if self.entryLabel:
            entryFrame = tk.Frame(frame)
            entryFrame.grid(row=2, column=0, sticky='new')
            tk.Label(entryFrame, text=self.entryLabel).grid(row=0, column=0,
                                                            sticky='nw', pady=3)
            tk.Entry(entryFrame,
                     textvariable=self.entryVar,
                     bg='white',
                     width=65).grid(row=0, column=1, sticky='nw', pady=3)

        frame.rowconfigure(treeRow, weight=1)

    def _addButton(self, frame, text, image, command):
        btn = tk.Label(frame, text=text, image=self.getImage(image),
                       compound=tk.LEFT, cursor='hand2')
        btn.bind('<Button-1>', command)
        btn.grid(row=0, column=self._col, sticky='nw',
                 padx=(0, 5), pady=5)
        self._col += 1

    def _fillToolbar(self, frame):
        """ Fill the toolbar frame with some buttons. """
        self._col = 0

        self._addButton(frame, 'Refresh', pwutils.Icon.ACTION_REFRESH,
                        self._actionRefresh)
        self._addButton(frame, 'Home', pwutils.Icon.HOME, self._actionHome)
        self._addButton(frame, 'Launch folder', pwutils.Icon.ROCKET,
                        self._actionLaunchFolder)
        self._addButton(frame, 'Working dir', pwutils.Icon.ACTION_BROWSE,
                        self._actionWorkingDir)
        self._addButton(frame, 'Up', pwutils.Icon.ARROW_UP, self._actionUp)

        # Add shortcuts
        self._addShortCuts(frame)

    def _addShortCuts(self, frame):
        """ Add shortcuts if available"""
        if self.shortCuts:
            for shortCut in self.shortCuts:
                self._addButton(frame,
                                shortCut.name,
                                shortCut.icon,
                                lambda e: self._goDir(shortCut.path))

    def _fillButtonsFrame(self, frame):
        """ Add button to the bottom frame if the selectMode
        is distinct from SELECT_NONE.
        """
        Button(frame, "Close", pwutils.Icon.BUTTON_CLOSE,
               command=self._close).grid(row=0, column=0, padx=(0, 5))
        if self.selectButton:
            HotButton(frame, self.selectButton, pwutils.Icon.BUTTON_SELECT,
                      command=self._select).grid(row=0, column=1)

    def _actionRefresh(self, e=None):
        self.tree.update()

    def _goDir(self, newDir):

        newDir = os.path.abspath(newDir)

        # Add a final "/" to the path: abspath is removing it except for "/"
        if not newDir.endswith(os.path.sep):
            newDir += os.path.sep

        self.pathVar.set(newDir)
        self.pathEntry.icursor(len(newDir))
        self.treeProvider.setDir(newDir)
        self.tree.update()
        self.tree.focus_set()

        itemKeyToFocus = PARENT_FOLDER
        if PARENT_FOLDER not in self.tree._objDict:
            itemKeyToFocus = self.tree.get_children()[0]

        # Focusing on a item, but nothing is selected 
        # Current dir remains in _lastSelected
        self._lastSelected = FileInfo(os.path.dirname(newDir),
                                      os.path.basename(newDir))
        self.tree.focus(itemKeyToFocus)

    def _actionUp(self, e=None):
        parentFolder = pwutils.getParentFolder(self.treeProvider.getDir())
        self._goDir(parentFolder)

    def _actionHome(self, e=None):
        self._goDir(pwutils.getHomePath())

    def _actionRoot(self, e=None):
        self._goDir("/")

    def _actionLaunchFolder(self, e=None):
        self._goDir(Config.SCIPION_CWD)

    def _actionWorkingDir(self, e=None):
        self._goDir(os.getcwd())

    def _itemDoubleClick(self, obj):
        if obj.isDir():
            self._goDir(obj.getPath())
        else:
            actions = self._provider.getObjectActions(obj)
            if actions:
                # actions[0] = first Action, [1] = the action callback
                actions[0][1]()

    def _itemKeyPressed(self, obj, e=None):

        if e.keysym in [pwutils.KEYSYM.RETURN]:
            self._itemDoubleClick(obj)
            return

        textToSearch = self._composeTextToSearch(e.char)

        # locate an item in starting with that letter.
        self._searchItem(textToSearch)

    def _composeTextToSearch(self, newChar):

        currentMiliseconds = time.time()

        if (self.previousSearchTS is not None) and \
                ((currentMiliseconds - self.previousSearchTS) < 0.3):
            newChar = self.previousSearch + newChar

        self.previousSearch = newChar
        self.previousSearchTS = currentMiliseconds

        return newChar

    def _searchItem(self, char):
        """ locate an item in starting with that letter."""
        try:
            self.tree.search(char)
        except Exception as e:
            # seems to raise an exception but selects things right.
            pass

    def _onEnterPath(self, e=None):
        path = os.path.abspath(self.pathVar.get())
        if os.path.exists(path):
            self._goDir(path)

        else:
            self.showInfo("Path '%s' does not exists. " % path)
            self.pathEntry.focus()

    def onClose(self):
        pass

    def onSelect(self, obj):
        print(obj, "type: ", type(obj))

    def _close(self, e=None):
        self.onClose()

    def _select(self, e=None):
        _lastSelected = self.getSelected()
        if _lastSelected is not None:
            self.onSelect(_lastSelected)
        else:
            print('Select a valid file/folder')

    def getEntryValue(self):
        return self.entryVar.get()

    def getCurrentDir(self):
        return self.treeProvider.getDir()


class ShortCut:
    """ Shortcuts to paths to be displayed in the file browser"""

    @staticmethod
    def factory(path, name, icon=None, toolTip=""):
        """ Factory method to create shortcuts"""
        return ShortCut(path, name, icon, toolTip)

    def __init__(self, path, name, icon=None, toolTip=""):
        self.path = path
        self.name = name
        self.icon = icon
        self.toolTip = toolTip


class BrowserWindow(gui.Window):
    """ Windows to hold a browser frame inside. """

    def __init__(self, title, master=None, **kwargs):
        if 'minsize' not in kwargs:
            kwargs['minsize'] = (800, 400)
        gui.Window.__init__(self, title, master, **kwargs)

    def setBrowser(self, browser, row=0, column=0):
        self.browser = browser
        browser.grid(row=row, column=column, sticky='news')
        self.itemConfig = browser.tree.itemConfig


STANDARD_IMAGE_EXTENSIONS = ['.png', '.jpg', '.jpeg']


def isStandardImage(filename):
    """ Check if a filename have an standard image extension. """
    fnLower = filename.lower()
    return any(fnLower.endswith(ext) for ext in STANDARD_IMAGE_EXTENSIONS)


class FileBrowserWindow(BrowserWindow):
    """ Windows to hold a file browser frame inside. """

    def __init__(self, title, master=None, path=None,
                 onSelect=None, shortCuts=None, **kwargs):
        BrowserWindow.__init__(self, title, master, **kwargs)
        self.registerHandlers()
        browser = FileBrowser(self.root, path,
                              showInfo=lambda msg: self.showInfo(msg, "Info"),
                              shortCuts=shortCuts,
                              **kwargs)
        if onSelect:
            def selected(obj):
                self.close()
                onSelect(obj)

            browser.onSelect = selected
        browser.onClose = self.close
        self.setBrowser(browser)

    def getEntryValue(self):
        return self.browser.getEntryValue()

    def getCurrentDir(self):
        return self.browser.getCurrentDir()

    def registerHandlers(self):
        register = FileTreeProvider.registerFileHandler  # shortcut

        register(TextFileHandler('file_text.gif'),
                 '.txt', '.log', '.out', '.err', '.stdout', '.stderr', '.emx',
                 '.json', '.xml', '.pam')
        register(TextFileHandler('file_python.gif'), '.py')
        register(TextFileHandler('file_java.gif'), '.java')
        register(SqlFileHandler(), '.sqlite', '.db')
        # register(MdFileHandler(), '.xmd', '.star', '.pos', '.ctfparam', '.doc')
        # register(ParticleFileHandler(),
        #          '.xmp', '.tif', '.tiff', '.spi', '.mrc', '.map', '.raw',
        #          '.inf', '.dm3', '.em', '.pif', '.psd', '.spe', '.ser', '.img',
        #          '.hed', *STANDARD_IMAGE_EXTENSIONS)
        # register(VolFileHandler(), '.vol')
        # register(StackHandler(), '.stk', '.mrcs', '.st', '.pif', '.dm4')
        # register(ChimeraHandler(), '.bild')



# **************************************************************************
# *
# * Authors:    Roberto Marabini       (roberto@cnb.csic.es)
#               J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import time
import threading
import uuid
from datetime import timedelta, datetime
from urllib.parse import urlencode
from urllib.request import build_opener, HTTPHandler

import pyworkflow.utils as pwutils
from . import config


class ProjectWorkflowNotifier(object):
    """ Implement different types of notifications about a given
    project. Currently, the protocols in a workflow will be sent.
    """

    def __init__(self, project):
        self.project = project

    def _getUuidFileName(self):
        return self.project.getLogPath("uuid.log")

    def _getDataFileName(self, fileName="data.log"):
        return self.project.getLogPath(fileName)

    def _getUuid(self):
        # Load (or create if not exits) a file
        # in the project Logs folder to store an unique
        # project identifier
        uuidFn = self._getUuidFileName()
        try:
            with open(uuidFn) as f:
                uuidValue = f.readline()
        except IOError:
            uuidValue = str(uuid.uuid4())
            with open(uuidFn, 'w') as f:
                f.write(uuidValue)

        return uuidValue

    def _modifiedBefore(self, seconds):
        """ Return True if the uuid.log file has been modified within a given
        number of seconds. """
        uuidFn = self._getUuidFileName()
        if not os.path.exists(uuidFn):
            return False
        mTime = datetime.fromtimestamp(os.path.getmtime(uuidFn))
        delta = datetime.now() - mTime

        return delta < timedelta(seconds=seconds)

    def _sendData(self, url, dataDict=None):
        try:
            # then connect to webserver a send json
            # set debuglevel=0 for no messages
            opener = build_opener(HTTPHandler(debuglevel=0))
            data = urlencode(dataDict).encode()
            content = opener.open(url, data=data).read()
            now = time.time()
            os.utime(self._getUuidFileName(), (now, now))
        except Exception as e:
            pass
            # print("Could not notify, maybe there is not internet connection.")

    def _dataModified(self, projectWorfklow):
        try:
            with open(self._getDataFileName()) as f:
                projectWorfklow2 = f.readline()
                if projectWorfklow2 == projectWorfklow:
                    return False
        except IOError:
            pass
        return True

    def notifyWorkflow(self):

        try:
            # check if environment exists otherwise abort
            if not pwutils.envVarOn('SCIPION_NOTIFY'):
                return

            # if project specifies not to send stats
            if self._isProjectMuted():
                return
            # Check the seconds range of the notify, by default one day
            seconds = int(os.environ.get('SCIPION_NOTIFY_SECONDS', '86400'))

            if self._modifiedBefore(seconds):  # notify not more than once a day
                return

            # INFO: now we are only sending the protocols names in the project.
            # We could pass namesOnly=False to get the full workflow template
            projectWorfklow = self.project.getProtocolsJson(namesOnly=True)

            # if list with workflow has not been altered do not sent it
            if not self._dataModified(projectWorfklow):
                return
            else:
                # For compatibility with version 1.0 check
                # if Log directory exists. If it does not
                # create it
                # TODO: REMOVE this check in scipion 1.3
                dataFile = self._getDataFileName()
                # create the folder of the file path if not exists
                pwutils.makeFilePath(dataFile)
                with open(dataFile, 'w') as f:
                    f.write(projectWorfklow)
            dataDict = {'project_uuid': self._getUuid(),
                        'project_workflow': projectWorfklow}

            urlName = os.environ.get('SCIPION_NOTIFY_URL',
                                     config.SCIPION_STATS_WORKFLOW_APP).strip()
            urlName += "addOrUpdateWorkflow/"
            t = threading.Thread(name="notifier", target=lambda: self._sendData(urlName, dataDict))
            t.start()  # will execute function in a separate thread
        except Exception as e:
            print("Can't report usage: ", e)

    def _isProjectMuted(self):
        """ Projects are muted if they come from tests, Since there is no flag for it
        we will assume that if the project name starts with Test it will be considered
        a test and therefore no statistics will be sent"""
        return os.path.basename(self.project.name).startswith("Test")

    def getEntryFromWebservice(self, uuid):
        if not pwutils.envVarOn('SCIPION_NOTIFY'):
            return
        urlName = os.environ.get('SCIPION_NOTIFY_URL').strip()
        # remove last directory
        urlName = os.path.split(urlName)[0]
        url = urlName + "/?project_uuid=" + uuid
        resultDict = self._sendData(url)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module define a Graph class and some utilities 
"""


class Node(object):
    """ A node inside the graph. """
    _count = 1

    def __init__(self, name=None, label=None):
        self._childs = []
        self._parents = []

        if name is None:
            name = str(self._count)
            self._count += 1
        self._name = name

        if label is None:
            label = name
        self._label = label

    def getName(self):
        return self._name

    def getLabel(self):
        return self._label

    def setLabel(self, newLabel):
        self._label = newLabel

    def isRoot(self):
        return len(self._parents) == 0

    def getChilds(self):
        return self._childs

    def addChild(self, *nodes):
        for n in nodes:
            if n not in self._childs:
                self._childs.append(n)
                n._parents.append(self)

    def getParent(self):
        """ Return the first parent in the list,
        if the node isRoot, None is returned.
        """
        if self.isRoot():
            return None

        return self._parents[0]

    def getParents(self):
        return self._parents

    def iterChilds(self):
        """ Iterate over all childs and subchilds. 
        Nodes can be visited more than once if have
        more than one parent.
        """
        for child in self._childs:
            for c in child.iterChilds():
                yield c

        yield self

    def countChilds(self, visitedNode=None, count=0):
        """ Iterate over all childs and subchilds.
        Nodes can be visited once
        """
        for child in self._childs:
            if child._name not in visitedNode:
                visitedNode[child._name] = True
            child.countChilds(visitedNode)
        return len(visitedNode)


    def iterChildsBreadth(self):
        """ Iter child nodes in a breadth-first order
        """
        for child in self._childs:
            yield child

        for child in self._childs:
            for child2 in child.iterChildsBreadth():
                yield child2

    def __str__(self):
        return "Node (id=%s, label=%s, root=%s)" % (self._name,
                                                    self.getLabel(),
                                                    self.isRoot())


class Graph(object):
    """Simple directed Graph class. 
    Implemented using adjacency lists.
    """

    def __init__(self, rootName='ROOT', root=None):
        self._nodes = []
        self._nodesDict = {}  # To retrieve nodes from name
        if root is None:
            self._root = self.createNode(rootName)
        else:
            self._root = root
            self._registerNode(root)

    def _registerNode(self, node):
        self._nodes.append(node)
        self._nodesDict[node.getName()] = node
        for child in node.getChilds():
            self._registerNode(child)

    def getRoot(self):
        return self._root

    def createNode(self, nodeName, nodeLabel=None):
        """ Add a node to the graph """
        node = Node(nodeName, nodeLabel)
        self._registerNode(node)

        return node

    def aliasNode(self, node, aliasName):
        """ Register an alias name for the node. """
        self._nodesDict[aliasName] = node

    def getNode(self, nodeName):
        return self._nodesDict.get(nodeName, None)

    def getNodes(self):
        return self._nodes

    def getRootNodes(self):
        """ Return all nodes that have no parent. """
        return [n for n in self._nodes if n.isRoot()]

    def printNodes(self):
        for node in self.getNodes():
            print("Node: ", node)
            print(" Childs: ", ','.join([c.getLabel()
                                         for c in node.getChilds()]))

    def _escape(self, label):
        return label.replace('.', '_').replace(' ', '_').replace('-', '_').replace('___', '_')

    def printDot(self, useId=True):
        """ If useId is True, use the node id for label the graph.
         If not, use the run name.
        """

        def getLabel(node):
            if useId:
                return node.getName()
            else:
                return node.getLabel()

        dotStr = "\ndigraph {\n"

        for node in self.getNodes():
            for child in node.getChilds():
                nodeLabel = self._escape(getLabel(node))
                childLabel = self._escape(getLabel(child))
                dotStr += "   %s -> %s;\n" % (nodeLabel, childLabel)
        dotStr += "}"
        print(dotStr)
        return dotStr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import json
import datetime as dt
from collections import OrderedDict
from configparser import ConfigParser

import pyworkflow.object as pwobj
from pyworkflow.mapper import SqliteMapper


class ProjectSettings(pwobj.Object):
    """ Store settings related to a project. """

    COLOR_MODE_STATUS = 0
    COLOR_MODE_LABELS = 1
    COLOR_MODE_AGE = 2
    COLOR_MODES = (COLOR_MODE_STATUS, COLOR_MODE_LABELS, COLOR_MODE_AGE)

    def __init__(self, confs={}, **kwargs):
        super().__init__(**kwargs)
        self.config = ProjectConfig()
        # Store the current view selected by the user
        self.currentProtocolsView = pwobj.String()
        # Store the color mode: 0= Status, 1=Labels, ...
        self.colorMode = pwobj.Integer(ProjectSettings.COLOR_MODE_STATUS)
        # Store graph nodes positions and other info
        self.nodeList = NodeConfigList()
        self.labelsList = LabelsList()  # Label list
        self.mapper = None  # This should be set when load, or write
        self.runsView = pwobj.Integer(1)  # by default the graph view
        self.readOnly = pwobj.Boolean(False)
        self.runSelection = pwobj.CsvList(int)  # Store selected runs
        self.dataSelection = pwobj.CsvList(int)  # Store selected runs
        # Some extra settings stored, now mainly used
        # from the webtools
        # Time when the project was created
        self.creationTime = pwobj.String(dt.datetime.now())
        # Number of days that this project is active
        # if None, the project will not expire
        # This is used in webtools where a limited time
        # is allowed for each project
        self.lifeTime = pwobj.Integer()
        # Set a disk quota for the project (in Gb)
        # if None, quota is unlimited
        self.diskQuota = pwobj.Integer()

    def commit(self):
        """ Commit changes made. """
        self.mapper.commit()

    def getRunsView(self):
        return self.runsView.get()

    def setRunsView(self, value):
        self.runsView.set(value)

    def getReadOnly(self):
        return self.readOnly.get()

    def setReadOnly(self, value):
        self.readOnly.set(value)

    def getCreationTime(self):
        return self.creationTime.datetime()

    def setCreationTime(self, value):
        self.creationTime.set(value)

    def getLifeTime(self):
        return self.lifeTime.get()

    def setLifeTime(self, value):
        self.lifeTime.set(value)

    def getConfig(self):
        return self.config

    def getProtocolView(self):
        return self.currentProtocolsView.get()

    def setProtocolView(self, protocolView):
        """ Set the new protocol Menu given its index.
        The new ProtocolMenu will be returned.
        """
        self.currentProtocolsView.set(protocolView)

    def getColorMode(self):
        return self.colorMode.get()

    def setColorMode(self, colorMode):
        """ Set the color mode to use when drawing the graph.
        """
        self.colorMode.set(colorMode)

    def statusColorMode(self):
        return self.getColorMode() == self.COLOR_MODE_STATUS

    def labelsColorMode(self):
        return self.getColorMode() == self.COLOR_MODE_LABELS

    def ageColorMode(self):
        return self.getColorMode() == self.COLOR_MODE_AGE

    def write(self, dbPath=None):
        self.setName('ProjectSettings')
        if dbPath is not None:
            self.mapper = SqliteMapper(dbPath, globals())
        else:
            if self.mapper is None:
                raise Exception("Can't write ProjectSettings without "
                                "mapper or dbPath")

        self.mapper.deleteAll()
        self.mapper.insert(self)
        self.mapper.commit()

    def getNodes(self):
        return self.nodeList

    def getNodeById(self, nodeId):
        return self.nodeList.getNode(nodeId)

    def addNode(self, nodeId, **kwargs):
        return self.nodeList.addNode(nodeId, **kwargs)

    def getLabels(self):
        return self.labelsList

    @classmethod
    def load(cls, dbPath):
        """ Load a ProjectSettings from dbPath. """
        classDict = dict(globals())
        classDict.update(pwobj.__dict__)
        mapper = SqliteMapper(dbPath, classDict)
        settingList = mapper.selectByClass('ProjectSettings')
        n = len(settingList)

        if n == 0:
            raise Exception("Can't load ProjectSettings from %s" % dbPath)
        elif n > 1:
            raise Exception("Only one ProjectSettings is expected in db, "
                            "found %d in %s" % (n, dbPath))

        settings = settingList[0]
        settings.mapper = mapper

        return settings


class ProjectConfig(pwobj.Object):
    """A simple base class to store ordered parameters"""

    def __init__(self, **args):
        super().__init__(**args)
        self.logo = pwobj.String('scipion_logo_small.gif')
        # Do not store this object, unless we implement some kind of
        # icon customization
        self._objDoStore = False


class MenuConfig(object):
    """Menu configuration in a tree fashion.
    Each menu can contains submenus.
    Leaf elements can contain actions"""

    def __init__(self, text=None, value=None,
                 icon=None, tag=None, **kwargs):
        """Constructor for the Menu config item.
        Arguments:
          text: text to be displayed
          value: internal value associated with the item.
          icon: display an icon with the item
          tag: put some tags to items
        **args: pass other options to base class.
        """
        self.text = text
        self.value = value
        self.icon = icon
        self.tag = tag
        self.shortCut = kwargs.get('shortCut', None)
        self.childs = pwobj.List()
        self.openItem = kwargs.get('openItem', False)

    def addSubMenu(self, text, value=None, **args):
        subMenu = type(self)(text, value, **args)
        self.childs.append(subMenu)
        return subMenu

    def __iter__(self):
        for v in self.childs:
            yield v

    def __len__(self):
        return len(self.childs)

    def isEmpty(self):
        return len(self.childs) == 0


class NodeConfig(pwobj.Scalar):
    """ Store Graph node information such as x, y. """

    def __init__(self, nodeId=0, x=None, y=None, selected=False, expanded=True,
                 visible=True):
        pwobj.Scalar.__init__(self)
        # Special node id 0 for project node
        self._values = {'id': nodeId,
                        'x': pwobj.Integer(x).get(0),
                        'y': pwobj.Integer(y).get(0),
                        'selected': selected,
                        'expanded': expanded,
                        'visible': pwobj.Boolean(visible).get(0),
                        'labels': []}

    def _convertValue(self, value):
        """Value should be a str with comma separated values
        or a list.
        """
        self._values = json.loads(value)

    def getObjValue(self):
        self._objValue = json.dumps(self._values)
        return self._objValue

    def get(self):
        return self.getObjValue()

    def getId(self):
        return self._values['id']

    def setX(self, x):
        self._values['x'] = x

    def getX(self):
        return self._values['x']

    def setY(self, y):
        self._values['y'] = y

    def getY(self):
        return self._values['y']

    def setPosition(self, x, y):
        self.setX(x)
        self.setY(y)

    def getPosition(self):
        return self.getX(), self.getY()

    def setSelected(self, selected):
        self._values['selected'] = selected

    def isSelected(self):
        return self._values['selected']

    def setExpanded(self, expanded):
        self._values['expanded'] = expanded

    def isExpanded(self):
        return self._values['expanded']

    def setVisible(self, visible):
        self._values['visible'] = visible

    def isVisible(self):
        if self._values.get('visible') is None:
            self._values['visible'] = True
        return self._values['visible']

    def setLabels(self, labels):
        self._values['labels'] = labels

    def getLabels(self):
        return self._values.get('labels', None)

    def __str__(self):
        return 'NodeConfig: %s' % self._values


class NodeConfigList(pwobj.List):
    """ Store all nodes information items and 
    also store a dictionary for quick access
    to nodes query.
    """

    def __init__(self):
        self._nodesDict = {}
        pwobj.List.__init__(self)

    def getNode(self, nodeId):
        return self._nodesDict.get(nodeId, None)

    def addNode(self, nodeId, **kwargs):
        node = NodeConfig(nodeId, **kwargs)
        self._nodesDict[node.getId()] = node
        self.append(node)
        return node

    def updateDict(self):
        self._nodesDict.clear()
        for node in self:
            self._nodesDict[node.getId()] = node

    def clear(self):
        pwobj.List.clear(self)
        self._nodesDict.clear()


class Label(pwobj.Scalar):
    """ Store Label information """

    def __init__(self, labelId=None, name='', color=None):
        pwobj.Scalar.__init__(self)
        # Special node id 0 for project node
        self._values = {'id': labelId,
                        'name': name,
                        'color': color}

    def _convertValue(self, value):
        """Value should be a str with comma separated values
        or a list.
        """
        self._values = json.loads(value)

    def getObjValue(self):
        self._objValue = json.dumps(self._values)
        return self._objValue

    def get(self):
        return self.getObjValue()

    def getId(self):
        return self._values['id']

    def getName(self):
        return self._values['name']

    def setName(self, newName):
        self._values['name'] = newName

    def setColor(self, color):
        self._values['color'] = color

    def getColor(self):
        return self._values.get('color', None)

    def __str__(self):
        return 'Label: %s' % self._values

    def __eq__(self, other):
        return self.getName() == other.getName()


class LabelsList(pwobj.List):
    """ Store all labels information"""

    def __init__(self):
        self._labelsDict = {}
        pwobj.List.__init__(self)

    def getLabel(self, name):
        return self._labelsDict.get(name, None)

    def addLabel(self, label):
        self._labelsDict[label.getName()] = label
        self.append(label)
        return label

    def updateDict(self):
        self._labelsDict.clear()
        for label in self:
            self._labelsDict[label.getName()] = label

    def deleteLabel(self, label):
        self._labelsDict.pop(label.getName())
        self.remove(label)

    def clear(self):
        pwobj.List.clear(self)
        self._labelDict.clear()



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     Pablo Conesa (pconesa@cnb.csic.es) [1]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from pyworkflow.project.project import Project
from unittest import TestCase
from unittest.mock import patch


class TestProject(TestCase):

    def test_fixlinks(self):
        """ Test fixlinks call."""

        with patch("pyworkflow.project.Project.getRuns") as getruns:

            getruns.return_value = [1]
            proj = Project("domain", "path")
            proj.fixLinks("foo")



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     Yaiza Rancel (yrancel@cnb.csic.es)
# *
# * Unidad de Bioinformatica of Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import sys
import os

import pyworkflow as pw
import pyworkflow.utils as pwutils
from pyworkflow.project import Manager, Project


def usage(error):
    print("""
    ERROR: %s

    Usage: python -m pyworkflow.project.scripts.stop project_name
        This script will stop all running protocols of the specified project.
        e.g.
        scipion python scripts/stop.py MyProject
    """ % error)
    sys.exit(1)


n = len(sys.argv)

if n > 2:
    usage("This script accepts 1 mandatory parameter: the project name")

projName = sys.argv[1]

path = pw.join('gui', 'no-tkinter')
sys.path.insert(1, path)

manager = Manager()

if not manager.hasProject(projName):
    usage("There is no project with this name: %s"
          % pwutils.red(projName))

# the project may be a soft link which may be unavailable to the cluster so get the real path
try:
    projectPath = os.readlink(manager.getProjectPath(projName))
except:
    projectPath = manager.getProjectPath(projName)

project = Project(path=projectPath)
project.load()

runs = project.getRuns()

# Now assuming that there is no dependencies between runs
# and the graph is lineal
for prot in runs:
    if prot.isActive():
        try:
            project.stopProtocol(prot)
        except:
            print("Couldn't stop protocol %s" % prot)



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

"""
Run or show the selected tests. Tests can be selected by giving
the "case", or by giving the paths and file pattern to use for
searching them.
"""
from os.path import basename
import argparse
from collections import OrderedDict

import pyworkflow.tests as pwtests
from pyworkflow import getTestsScript, SCIPION_TESTS_CMD, Config

from pyworkflow.tests import *

MODULE = 0
CLASS = 1
TEST = 2


class Tester:
    def main(self, args=None):
        print("Running tests....")

        # Trigger plugin's variable definition
        Config.getDomain().getPlugins()

        parser = argparse.ArgumentParser(prog=self.getTestsCommand(), description=__doc__)
        g = parser.add_mutually_exclusive_group()
        g.add_argument('--run', action='store_true', help='run the selected tests')
        g.add_argument('--show', action='store_true', help='show available tests',
                       default=True)

        add = parser.add_argument  # shortcut

        add('--pattern', default='test*.py',
            help='pattern for the files that will be used in the tests')
        add('--grep', default=None, nargs='+',
            help='only show/run tests containing the provided words')
        add('--skip', default=None, nargs='+',
            help='skip tests that contains these words')
        add('--log', default=None, nargs='?',
            help="Generate logs files with the output of each test.")
        add('--mode', default='classes', choices=['modules', 'classes', 'onlyclasses', 'all'],
            help='how much detail to give in show mode')
        add('tests', metavar='TEST', nargs='*',
            help='test case from string identifier (module, class or callable)')
        args = parser.parse_args(args)

        if not args.run and not args.show and not args.tests:
            sys.exit(parser.format_help())

        testsDict = OrderedDict()
        testLoader = unittest.defaultTestLoader

        if args.tests:
            # In this case the tests are passed as argument.
            # The full name of the test will be used to load it.
            testsDict['tests'] = unittest.TestSuite()
            for t in args.tests:
                try:
                    testsDict['tests'].addTests(testLoader.loadTestsFromName(t))
                except Exception as e:
                    print('Cannot load test %s -- skipping' % t)
                    import traceback
                    traceback.print_exc()
        else:
            # In this other case, we will load the test available
            # from pyworkflow and the other plugins
            # self.paths = [('pyworkflow', os.path.dirname(os.path.dirname(pw.__file__)))]
            self.paths = []
            for name, plugin in pw.Config.getDomain().getPlugins().items():
                self.paths.append((name, os.path.dirname(plugin.__path__[0])))
            for k, p in self.paths:
                testPath = os.path.join(p, k, 'tests')
                if os.path.exists(testPath):
                    testsDict[k] = testLoader.discover(testPath,
                                                       pattern=args.pattern,
                                                       top_level_dir=p)

        self.grep = [g.lower() for g in args.grep] if args.grep else []
        self.skip = args.skip
        self.mode = args.mode
        self.log = args.log

        if args.tests:
            self.runSingleTest(testsDict['tests'])

        elif args.run:
            for moduleName, tests in testsDict.items():
                print(pwutils.cyan(">>>> %s" % moduleName))
                self.runTests(moduleName, tests)

        elif args.grep:
            pattern = args.grep[0]
            for moduleName, tests in testsDict.items():
                self.printTests(pattern, tests)

        else:
            for moduleName, tests in testsDict.items():
                if self._match(moduleName):
                    print(pwutils.cyan(">>>> %s" % moduleName))
                    self.printTests(moduleName, tests)

    def _match(self, itemName):
        itemLower = itemName.lower()
        grep = (not self.grep or
                all(g.lower() in itemLower for g in self.grep))
        skip = (self.skip and
                any(g.lower() in itemLower for g in self.skip))

        return grep and not skip

    def __iterTests(self, test):
        """ Recursively iterate over a testsuite. """
        print("__iterTests: %s, is-suite: %s" % (str(test.__class__),
                                                 isinstance(test, unittest.TestSuite)))

        if isinstance(test, unittest.TestSuite):
            for t in test:
                self.__iterTests(t)
        else:
            yield test

    def _visitTests(self, moduleName, tests, newItemCallback):
        """ Show the list of tests available """
        mode = self.mode

        assert mode in ['modules', 'classes', 'onlyclasses', 'all'], 'Unknown mode %s' % mode

        # First flatten the list of tests.
        # testsFlat = list(iter(self.__iterTests(tests)))

        testsFlat = []
        toCheck = [t for t in tests]

        while toCheck:
            test = toCheck.pop()
            if isinstance(test, unittest.TestSuite):
                toCheck += [t for t in test]
            elif test not in testsFlat:
                testsFlat.append(test)

        # Follow the flattened list of tests and show the module, class
        # and name, in a nice way.
        lastClass = None
        lastModule = None
        if testsFlat:
            for t in testsFlat:

                testModuleName, className, testName = t.id().rsplit('.', 2)

                # If there is a failure loading the test, show it
                errorStr = 'unittest.loader._FailedTest.'
                if testModuleName.startswith(errorStr):
                    newName = t.id().replace(errorStr, '')
                    if self._match(newName):
                        print(pwutils.red('Error loading the test. Please, run the test for more information:'),
                              newName)
                    continue

                if testModuleName != lastModule:
                    lastModule = testModuleName
                    if mode != 'onlyclasses':
                        newItemCallback(MODULE, "%s" % testModuleName)

                if mode in ['classes', 'onlyclasses', 'all'] and className != lastClass:
                    lastClass = className
                    newItemCallback(CLASS, "%s.%s"
                                    % (testModuleName, className))

                if mode == 'all':
                    newItemCallback(TEST, "%s.%s.%s"
                                    % (testModuleName, className, testName))
        else:
            if not self.grep:
                print(pwutils.green(' The plugin does not have any test'))

    def _printNewItem(self, itemType, itemName):
        if self._match(itemName):
            spaces = (itemType * 2) * ' '
            print("%s %s %s" % (spaces, self.getTestsCommand(), itemName))

    def getTestsCommand(self):
        return os.environ.get(SCIPION_TESTS_CMD, getTestsScript())

    def printTests(self, moduleName, tests):
        self._visitTests(moduleName, tests, self._printNewItem)

    def _logTest(self, cmd, runTime, result, logFile):
        with open(self.testLog, "r+") as f:
            lines = f.readlines()
            f.seek(0)
            for l in lines:
                if '<!-- LAST_ROW -->' in l:
                    rowStr = "<tr><td>%s</td><td>%s</td><td>%s</td><td>%s</td><td>%s</td></tr>"
                    if result:  # != 0 means failed in os.system
                        resultStr = '<font color="red">[FAILED]</font>'
                    else:
                        resultStr = '<font color="green">[SUCCEED]</font>'
                    logStr = '<a href="file://%s">%s</a>' % (logFile, basename(logFile))

                    f.write(rowStr % (self.testCount, cmd, runTime, resultStr, logStr))
                    f.write('\n')
                if self.headerPrefix in l:
                    f.write(self.headerPrefix + self.testTimer.getToc() + '</h3>\n')
                else:
                    f.write(l)
            f.close()

    def _runNewItem(self, itemType, itemName):
        if self._match(itemName):
            spaces = (itemType * 2) * ' '
            script = getTestsScript()
            cmd = "%s %s %s %s" % (pw.PYTHON, script, spaces, itemName)
            run = ((itemType == MODULE and self.mode == 'modules') or
                   (itemType == CLASS and self.mode in ('classes', 'onlyclasses')) or
                   (itemType == TEST and self.mode == 'all'))
            if run:
                if self.log:
                    logFile = join(self.testsDir, '%s.txt' % itemName)
                    cmdFull = cmd + " > %s 2>&1" % logFile
                else:
                    logFile = ''
                    cmdFull = cmd

                print(pwutils.green(cmdFull))
                t = pwutils.Timer()
                t.tic()
                self.testCount += 1
                result = os.system(cmdFull)
                if self.log:
                    self._logTest(cmd, t.getToc(), result, logFile)

    def runTests(self, moduleName, tests):
        self.testCount = 0

        if self.log:
            self.testsDir = join(pw.Config.SCIPION_USER_DATA, 'Tests', self.log)
            pwutils.cleanPath(self.testsDir)
            pwutils.makePath(self.testsDir)
            self.testLog = join(self.testsDir, 'tests.html')
            self.testTimer = pwutils.Timer()
            self.testTimer.tic()
            self.headerPrefix = '<h3>Test results (%s) Duration: ' % pwutils.prettyTime()
            f = open(self.testLog, 'w')
            f.write("""<!DOCTYPE html>
    <html>
    <body>
    """)
            f.write(self.headerPrefix + '</h3>')
            f.write("""    
     <table style="width:100%" border="1">
      <tr>
        <th>#</th>
        <th>Command</th>
        <th>Time</th>
        <th>Result</th>
        <th>Log file</th>
      </tr>
    <!-- LAST_ROW -->
    </table> 
    
    </body>
    </html>""")

            f.close()
        self._visitTests(moduleName, tests, self._runNewItem)

        if self.log:
            print("\n\nOpen results in your browser: \nfile:///%s"
                  % self.testLog)

    def runSingleTest(self, tests):
        result = pwtests.GTestResult()
        tests.run(result)
        result.doReport()
        resultPassed = result.numberTests - result.testFailed
        sys.exit(1 if result.testFailed > 0 or resultPassed == 0 else 0)


if __name__ == '__main__':
    Tester().main()
